/*
 * Author: Dušan Kolář
 */

#include "pcx.h"

#include <malloc.h>
#include <string.h>
#include <stdbool.h>


#define HEAD_SIZE         4
#define PROTOCOL_VERSION  1
#define MAX_TOKEN_LEN     8
#define PAYLOAD_MARKER    0xff


#define SWAP_TYPE(WIDTH) typedef union u_swap ## WIDTH {\
    uint ## WIDTH ## _t value;\
    uint8_t  bytes[WIDTH/8];\
} swap ## WIDTH ## _t;

SWAP_TYPE(64)
SWAP_TYPE(32)
SWAP_TYPE(16)


typedef struct s_header {
    uint8_t tkl:4, type:2, version:2;
    uint8_t code;
    swap16_t id;
} header_t;


#define OPT_LIST_TYPE(TYPE,VALUE) typedef struct s_ ## TYPE ##_list {\
    TYPE ## _opt_t              *VALUE;\
    struct s_ ## TYPE ##_list   *next;\
} TYPE ## _list_t;

OPT_LIST_TYPE(str,string)
OPT_LIST_TYPE(opaq,opaque)
OPT_LIST_TYPE(uint,uint_val)


typedef struct s_empty_list {
    uint16_t             value;
    struct s_empty_list *next;
} empty_list_t;

typedef union u_opt_header {
    uint8_t  opt_hdr;
    struct { uint8_t length:4, delta:4; } x;
} opt_header_t;

typedef enum e_option_type {
    Opaque, String, Empty, UInt, Skip
} option_type_t;

// .................................................................
// local function declaration

static void free_string_list(str_list_t *list);
static void free_opaque_list(opaq_list_t *list);
static void free_uint_val_list(uint_list_t *list);

static void free_empty_list(empty_list_t *list);


static str_opt_t  **mk_string_array(uint16_t size, str_list_t *list);
static opaq_opt_t **mk_opaque_array(uint16_t size, opaq_list_t *list);
static uint_opt_t **mk_uint_val_array(uint16_t size, uint_list_t *list);

static uint16_t    *mk_empty_array(uint16_t size, empty_list_t *list);


static coap_data_t *create_data(unsigned int type, uint8_t code, uint16_t id);

static int32_t  get_option_delta(uint8_t opt_hdr, uint8_t *input, uint32_t size);
static int32_t  get_option_length(uint8_t opt_hdr, uint8_t *input, uint32_t size, uint32_t delta);
static uint32_t get_opt_hdr_size(int32_t delta, int32_t length);

static option_type_t get_option_type(uint16_t option_number);

static parse_error_t parse_empty (uint16_t *empty_num, empty_list_t **empty_list, uint16_t option_number);
static parse_error_t parse_string(uint16_t *str_num,   str_list_t   **str_list,   uint16_t option_number,
                                  uint8_t *input, uint16_t length);
static parse_error_t parse_opaque(uint16_t *opaq_num, opaq_list_t **opaq_list, uint16_t option_number,
                                  uint8_t *input, uint16_t length);
static parse_error_t parse_uint(uint16_t *uint_num, uint_list_t **uint_list, uint16_t option_number,
                                  uint8_t *input, uint16_t length);

// .................................................................
// MACROS representing chain of void operations


#define FREE_LISTS() free_string_list(str_list);free_opaque_list(opaq_list);\
                     free_uint_val_list(uint_list);free_empty_list(empty_list)

// .................................................................
// ------------------------ GLOBAL FUNCTIONS ------------------------


coap_data_t *parse_coap(uint8_t *input, uint32_t size, parse_error_t *error) {

    // test if the packet is sufficiently long
    if (size < HEAD_SIZE) {
        *error = PE_TOO_SHORT;
        return (NULL);
    }

    header_t *h = (header_t *) input;

    // test protocol version
    if (h->version != PROTOCOL_VERSION) {
        *error = PE_EXPECTED_VERSION_01;
        return (NULL);
    }

    // test token length value
    if (h->tkl > MAX_TOKEN_LEN) {
        *error = PE_TOKEN_LENGTH_SIZE;
        return (NULL);
    }

    input += HEAD_SIZE;  // head parsed
    size  -= HEAD_SIZE;
    
    uint8_t *token = NULL;

    // skip token if present
    if (h->tkl > 0) {
        if (h->tkl > size) {
            *error = PE_OUT_OF_DATA;
            return (NULL);
        }
        token  = input;
        input += h->tkl; // move beyond token
        size  -= h->tkl;
    }

    uint32_t      payload_length = 0;
    uint8_t      *payload        = NULL;
    uint16_t      str_num        = 0;
    str_list_t   *str_list       = NULL;
    uint16_t      opaq_num       = 0;
    opaq_list_t  *opaq_list      = NULL;
    uint16_t      uint_num       = 0;
    uint_list_t  *uint_list      = NULL;
    uint16_t      empty_num      = 0;
    empty_list_t *empty_list     = NULL;
    uint16_t      option_number  = 0;
    
    // parse options and payload
    while (size>0) {
        uint8_t opt_hdr = *input;
        printf("optHdr %2x\n",opt_hdr);
        
        if (opt_hdr == PAYLOAD_MARKER) {
            --size;
            ++input;
            if (size==0) {
                FREE_LISTS();
                *error = PE_ZERO_LENGTH_PAYLOAD;
                return(NULL);
            }
            payload_length = size;
            payload = input;
            break;
            
        } else { // possibly option
            if ((opt_hdr & 0xf0) == 0xf0 || (opt_hdr & 0x0f) == 0x0f) {
                FREE_LISTS();
                *error = PE_WRONG_PAYLOAD_MARKER;
                return(NULL);
            }
            int32_t delta = get_option_delta(opt_hdr,input,size);
            if (delta<0) {
                FREE_LISTS();
                *error = PE_OUT_OF_DATA;
                return(NULL);
            }
            int32_t length = get_option_length(opt_hdr,input,size,delta);
            if (length<0) {
                FREE_LISTS();
                *error = PE_OUT_OF_DATA;
                return(NULL);
            }
            uint32_t opt_hdr_size = get_opt_hdr_size(delta, length);
            size -= opt_hdr_size;
            input += opt_hdr_size;
            printf("l %d\ns %d\nd %d\n",length,size,delta);
            
            // are we still in the packet ?
            if (length > size) {
                FREE_LISTS();
                *error = PE_OUT_OF_DATA;
                return(NULL);
            }
            
            option_number += delta;
            option_type_t opt_type = get_option_type(option_number);
            parse_error_t opt_res;
            switch (opt_type) {
                case String: {
                    if ((opt_res=parse_string(&str_num, &str_list, option_number, input, length))!=PE_OK) {
                        FREE_LISTS();
                        *error = opt_res;
                        return(NULL);
                    }
                    break;
                }
                case Opaque: {
                    if ((opt_res=parse_opaque(&opaq_num, &opaq_list, option_number, input, length))!=PE_OK) {
                        FREE_LISTS();
                        *error = opt_res;
                        return(NULL);
                    }
                    break;
                }
                case UInt: {
                    if ((opt_res=parse_uint(&uint_num, &uint_list, option_number, input, length))!=PE_OK) {
                        FREE_LISTS();
                        *error = opt_res;
                        return(NULL);
                    }
                    break;
                }
                case Empty: {
                    if (length != 0) {
                        FREE_LISTS();
                        *error = PE_OPTION_OUT_OF_BOUNDS;
                        return(NULL);
                    }
                    if ((opt_res=parse_empty(&empty_num, &empty_list, option_number))!=PE_OK) {
                        FREE_LISTS();
                        *error = opt_res;
                        return(NULL);
                    }
                    break;
                }
                case Skip: {
                    break;
                }
            } // switch
            input += length;
            size  -= length;
        }
    }  // while size > 0


    // save message ID correctly on Intel arch.
    swap16_t message_id;
    message_id.bytes[0] = h->id.bytes[1];
    message_id.bytes[1] = h->id.bytes[0];

    coap_data_t *res = create_data(h->type, h->code, message_id.value);
    
    if (res==NULL) {
        FREE_LISTS();
        *error = PE_OUT_OF_MEMORY;
        return(NULL);
    }
    
    // set token value if present
    if (h->tkl > 0) {
        unsigned max = res->tkl = h->tkl;
        for (unsigned i=0; i<max; ++i) {
            res->token[i] = token[i];
        }
    }
    
    // set payload if present
    if (payload_length > 0) {
        res->payload = malloc(payload_length);
        if (res->payload == NULL) {
            FREE_LISTS();
            free(res);
            *error = PE_OUT_OF_MEMORY;
            return(NULL);
        }
        res->payload_length = payload_length;
        (void)memcpy(res->payload, payload, payload_length);
    }

    res->options_num = str_num + opaq_num + uint_num + empty_num;
    
    if (str_num > 0) {
        res->str_num = str_num;
        if ((res->strings = mk_string_array(str_num,str_list)) == NULL) {
            free_coap_data(res);
            FREE_LISTS();
            *error = PE_OUT_OF_MEMORY;
            return(NULL);
        }
        str_list = NULL;
    }
    
    if (opaq_num > 0) {
        res->opaq_num = opaq_num;
        if ((res->opaques = mk_opaque_array(opaq_num,opaq_list)) == NULL) {
            free_coap_data(res);
            free_opaque_list(opaq_list);
            free_uint_val_list(uint_list);
            free_empty_list(empty_list);
            *error = PE_OUT_OF_MEMORY;
            return(NULL);
        }
        opaq_list = NULL;
    }

    if (uint_num > 0) {
        res->uint_num = uint_num;
        if ((res->uints = mk_uint_val_array(uint_num,uint_list)) == NULL) {
            free_coap_data(res);
            free_uint_val_list(uint_list);
            free_empty_list(empty_list);
            *error = PE_OUT_OF_MEMORY;
            return(NULL);
        }
        uint_list = NULL;
    }
    
    if (empty_num > 0) {
        res->empty_num = empty_num;
        if ((res->emptys = mk_empty_array(empty_num,empty_list)) == NULL) {
            free_coap_data(res);
            free_empty_list(empty_list);
            *error = PE_OUT_OF_MEMORY;
            return(NULL);
        }
        uint_list = NULL;
    }
    
    *error = PE_OK;
    return (res);
}


// .................................................................


void free_coap_data(coap_data_t *data) {
    if (data == NULL) return;
    // payload
    if (data->payload!=NULL && data->payload_length>0) {
        free((void*)data->payload);
    }
    // string data
    if (data->strings!=NULL && data->str_num>0) {
        uint16_t locnum = data->str_num;
        for (uint16_t i=0; i<locnum; ++i) {
            free((void*)data->strings[i]);
        }
        free((void*)data->strings);
    }
    // opaque data
    if (data->opaques!=NULL && data->opaq_num>0) {
        uint16_t locnum = data->opaq_num;
        for (uint16_t i=0; i<locnum; ++i) {
            free((void*)data->opaques[i]);
        }
        free((void*)data->opaques);
    }
    // uint data
    if (data->uints!=NULL && data->uint_num>0) {
        uint16_t locnum = data->uint_num;
        for (uint16_t i=0; i<locnum; ++i) {
            free((void*)data->uints[i]);
        }
        free((void*)data->uints);
    }
    // empty data
    if (data->emptys!=NULL && data->empty_num>0) {
        free((void*)data->emptys);
    }
    free((void*)data);
} // free_coap_data


// .................................................................
// ------------------------ LOCAL FUNCTIONS ------------------------


#define FREE_PTR_LIST(TYPE,VALUE) static void free_ ## VALUE ## _list(TYPE *list) {\
    while (list != NULL) {\
        TYPE *tmp = list;\
        free((void*)list->VALUE);\
        list = list->next;\
        free((void*)tmp);\
    }\
}

FREE_PTR_LIST(str_list_t,string)
FREE_PTR_LIST(opaq_list_t,opaque)
FREE_PTR_LIST(uint_list_t,uint_val)

// .................................................................

static void free_empty_list(empty_list_t *list) {
    while (list != NULL) {
        empty_list_t *tmp = list;
        list = list->next;
        free((void*)tmp);
    }
}


// .................................................................

#define MK_PTR_ARRAY(TYPE,LIST_TYPE,VALUE) static TYPE **mk_ ## VALUE ## _array(uint16_t size, LIST_TYPE *list) {\
    TYPE **res = malloc(size * sizeof(TYPE*));\
    if (res == NULL) return(NULL);\
    for(uint16_t i=0; i<size; ++i) {\
        LIST_TYPE *tmp = list;\
        res[i] = list->VALUE;\
        list = list->next;\
        free((void*)tmp);\
    }\
    return(res);\
}


MK_PTR_ARRAY(str_opt_t,str_list_t,string)
MK_PTR_ARRAY(opaq_opt_t,opaq_list_t,opaque)
MK_PTR_ARRAY(uint_opt_t,uint_list_t,uint_val)

// .................................................................

static uint16_t *mk_empty_array(uint16_t size, empty_list_t *list) {
    uint16_t *res = malloc(size * sizeof(uint16_t));
    if (res == NULL) return(NULL);
    for(uint16_t i=0; i<size; ++i) {
        empty_list_t *tmp = list;
        res[i] = list->value;
        list = list->next;
        free((void*)tmp);
    }
    return(res);
}


// .................................................................

static coap_data_t *create_data(unsigned int type, uint8_t code, uint16_t id) {
    coap_data_t *res = malloc(sizeof(coap_data_t));
    if (res == NULL) return(NULL);
    res->type = type;
    res->code = code;
    res->id   = id;
    
    res->tkl            = 0;
    res->options_num    = 0;
    res->str_num        = 0;
    res->opaq_num       = 0;
    res->empty_num      = 0;
    res->uint_num       = 0;
    res->payload_length = 0;
    
    res->strings = NULL;
    res->opaques = NULL;
    res->emptys  = NULL;
    res->payload = NULL;
    res->uints   = NULL;
    return(res);
}

// .................................................................

static int32_t get_option_delta(uint8_t opt_hdr, uint8_t *input, uint32_t size) {
    opt_header_t header;
    header.opt_hdr = opt_hdr;
    printf("del %d\n",header.x.delta);
    
    if (header.x.delta < 13) {
        return((int32_t)header.x.delta);
        
    } else if (header.x.delta == 13) {
        if (size < 2) return(-1);
        return(input[1]+13);
        
    } else {
        // header.delta == 14, value 15 is tested elsewhere and reserved
        if (size < 3) return(-2);
        swap16_t delta;
        delta.bytes[0] = input[2];
        delta.bytes[1] = input[1];
        return(delta.value+269);
    }
}

// .................................................................

static int32_t get_option_length(uint8_t opt_hdr, uint8_t *input, uint32_t size, uint32_t delta) {
    opt_header_t header;
    header.opt_hdr = opt_hdr;
    
    if (header.x.length < 13) {
        return((int32_t)header.x.length);

    } else {
        uint32_t offset = 0;
        
        if (delta>=269) offset = 2;
        else if (delta>=13) offset=1;
        
        if (header.x.length == 13) {
            if (size < (offset+2)) return(-1);
            return(input[1+offset]+13);
            
        } else {
            // header.length == 14, value 15 is tested elsewhere and reserved
            if (size < (offset+3)) return(-2);
            swap16_t length;
            length.bytes[0] = input[2+offset];
            length.bytes[1] = input[1+offset];
            return(length.value+269);
        }
    }
}

// .................................................................

static uint32_t get_opt_hdr_size(int32_t delta, int32_t length) {
    uint32_t offset = 1;
        
    if (delta>=269) offset += 2;
    else if (delta>=13) offset += 1;
    
    if (length>=269) offset += 2;
    else if (length>=13) offset += 1;
    
    return(offset);
}

// .................................................................

/*
+--------+------------------+-----------+
| Number | Name             | Reference |
+--------+------------------+-----------+
|      0 | (Reserved)       | [RFC7252] |
|      1 | If-Match         | [RFC7252] |
|      3 | Uri-Host         | [RFC7252] |
|      4 | ETag             | [RFC7252] |
|      5 | If-None-Match    | [RFC7252] |
|      7 | Uri-Port         | [RFC7252] |
|      8 | Location-Path    | [RFC7252] |
|     11 | Uri-Path         | [RFC7252] |
|     12 | Content-Format   | [RFC7252] |
|     14 | Max-Age          | [RFC7252] |
|     15 | Uri-Query        | [RFC7252] |
|     17 | Accept           | [RFC7252] |
|     20 | Location-Query   | [RFC7252] |
|     35 | Proxy-Uri        | [RFC7252] |
|     39 | Proxy-Scheme     | [RFC7252] |
|     60 | Size1            | [RFC7252] |
|    128 | (Reserved)       | [RFC7252] |
|    132 | (Reserved)       | [RFC7252] |
|    136 | (Reserved)       | [RFC7252] |
|    140 | (Reserved)       | [RFC7252] |
+--------+------------------+-----------+
+-----+---+---+---+---+----------------+--------+--------+----------+
| No. | C | U | N | R | Name           | Format | Length | Default  |
+-----+---+---+---+---+----------------+--------+--------+----------+
|   1 | x |   |   | x | If-Match       | opaque | 0-8    | (none)   |
|   3 | x | x | - |   | Uri-Host       | string | 1-255  | (see     |
|     |   |   |   |   |                |        |        | below)   |
|   4 |   |   |   | x | ETag           | opaque | 1-8    | (none)   |
|   5 | x |   |   |   | If-None-Match  | empty  | 0      | (none)   |
|   7 | x | x | - |   | Uri-Port       | uint   | 0-2    | (see     |
|     |   |   |   |   |                |        |        | below)   |
|   8 |   |   |   | x | Location-Path  | string | 0-255  | (none)   |
|  11 | x | x | - | x | Uri-Path       | string | 0-255  | (none)   |
|  12 |   |   |   |   | Content-Format | uint   | 0-2    | (none)   |
|  14 |   | x | - |   | Max-Age        | uint   | 0-4    | 60       |
|  15 | x | x | - | x | Uri-Query      | string | 0-255  | (none)   |
|  17 | x |   |   |   | Accept         | uint   | 0-2    | (none)   |
|  20 |   |   |   | x | Location-Query | string | 0-255  | (none)   |
|  35 | x | x | - |   | Proxy-Uri      | string | 1-1034 | (none)   |
|  39 | x | x | - |   | Proxy-Scheme   | string | 1-255  | (none)   |
|  60 |   |   | x |   | Size1          | uint   | 0-4    | (none)   |
+-----+---+---+---+---+----------------+--------+--------+----------+
*/

#define IF_MATCH            1
#define IF_MATCH_TYPE       Opaque
#define IF_MATCH_MIN        0
#define IF_MATCH_MAX        8

#define URI_HOST            3
#define URI_HOST_TYPE       String
#define URI_HOST_MIN        1
#define URI_HOST_MAX        255

#define ETAG                4
#define ETAG_TYPE           Opaque
#define ETAG_MIN            1
#define ETAG_MAX            8

#define IF_NONE_MATCH       5
#define IF_NONE_MATCH_TYPE  Empty
#define IF_NONE_MATCH_MIN   0
#define IF_NONE_MATCH_MAX   0

#define URI_PORT            7
#define URI_PORT_TYPE       UInt
#define URI_PORT_MIN        0
#define URI_PORT_MAX        2

#define LOCATION_PATH       8
#define LOCATION_PATH_TYPE  String
#define LOCATION_PATH_MIN   0
#define LOCATION_PATH_MAX   255

#define URI_PATH            11
#define URI_PATH_TYPE       String
#define URI_PATH_MIN        0
#define URI_PATH_MAX        255

#define CONTENT_FORMAT      12
#define CONTENT_FORMAT_TYPE UInt
#define CONTENT_FORMAT_MIN  0
#define CONTENT_FORMAT_MAX  2

#define MAX_AGE             14
#define MAX_AGE_TYPE        UInt
#define MAX_AGE_MIN         0
#define MAX_AGE_MAX         4

#define URI_QUERY           15
#define URI_QUERY_TYPE      String
#define URI_QUERY_MIN       0
#define URI_QUERY_MAX       255

#define ACCEPT              17
#define ACCEPT_TYPE         UInt
#define ACCEPT_MIN          0
#define ACCEPT_MAX          2

#define LOCATION_QUERY      20
#define LOCATION_QUERY_TYPE String
#define LOCATION_QUERY_MIN  0
#define LOCATION_QUERY_MAX  255

#define PROXY_URI           35
#define PROXY_URI_TYPE      String
#define PROXY_URI_MIN       1
#define PROXY_URI_MAX       1034

#define PROXY_SCHEME        39
#define PROXY_SCHEME_TYPE   String
#define PROXY_SCHEME_MIN    1
#define PROXY_SCHEME_MAX    255

#define SIZE1               60
#define SIZE1_TYPE          UInt
#define SIZE1_MIN           0
#define SIZE1_MAX           4

static option_type_t get_option_type(uint16_t option_number) {
    switch (option_number) {
        case IF_MATCH: {
            return(IF_MATCH_TYPE);
        }
        case URI_HOST: {
            return(URI_HOST_TYPE);
        }
        case ETAG: {
            return(ETAG_TYPE);
        }
        case IF_NONE_MATCH: {
            return(IF_NONE_MATCH_TYPE);
        }
        case URI_PORT: {
            return(URI_PORT_TYPE);
        }
        case LOCATION_PATH: {
            return(LOCATION_PATH_TYPE);
        }
        case URI_PATH: {
            return(URI_PATH_TYPE);
        }
        case CONTENT_FORMAT: {
            return(CONTENT_FORMAT_TYPE);
        }
        case MAX_AGE: {
            return(MAX_AGE_TYPE);
        }
        case URI_QUERY: {
            return(URI_QUERY_TYPE);
        }
        case ACCEPT: {
            return(ACCEPT_TYPE);
        }
        case LOCATION_QUERY: {
            return(LOCATION_QUERY_TYPE);
        }
        case PROXY_URI: {
            return(PROXY_URI_TYPE);
        }
        case PROXY_SCHEME: {
            return(PROXY_SCHEME_TYPE);
        }
        case SIZE1: {
            return(SIZE1_TYPE);
        }
    } // switch
    return(Skip);
}

// .................................................................

static bool option_length_ok(uint16_t option_number, uint16_t length) {
    switch (option_number) {
        case IF_MATCH: {
            return(IF_MATCH_MIN<=length && length<=IF_MATCH_MAX);
        }
        case URI_HOST: {
            return(URI_HOST_MIN<=length && length<=URI_HOST_MAX);
        }
        case ETAG: {
            return(ETAG_MIN<=length && length<=ETAG_MAX);
        }
        case IF_NONE_MATCH: {
            return(IF_NONE_MATCH_MIN<=length && length<=IF_NONE_MATCH_MAX);
        }
        case URI_PORT: {
            return(URI_PORT_MIN<=length && length<=URI_PORT_MAX);
        }
        case LOCATION_PATH: {
            return(LOCATION_PATH_MIN<=length && length<=LOCATION_PATH_MAX);
        }
        case URI_PATH: {
            return(URI_PATH_MIN<=length && length<=URI_PATH_MAX);
        }
        case CONTENT_FORMAT: {
            return(CONTENT_FORMAT_MIN<=length && length<=CONTENT_FORMAT_MAX);
        }
        case MAX_AGE: {
            return(MAX_AGE_MIN<=length && length<=MAX_AGE_MAX);
        }
        case URI_QUERY: {
            return(URI_QUERY_MIN<=length && length<=URI_QUERY_MAX);
        }
        case ACCEPT: {
            return(ACCEPT_MIN<=length && length<=ACCEPT_MAX);
        }
        case LOCATION_QUERY: {
            return(LOCATION_QUERY_MIN<=length && length<=LOCATION_QUERY_MAX);
        }
        case PROXY_URI: {
            return(PROXY_URI_MIN<=length && length<=PROXY_URI_MAX);
        }
        case PROXY_SCHEME: {
            return(PROXY_SCHEME_MIN<=length && length<=PROXY_SCHEME_MAX);
        }
        case SIZE1: {
            return(SIZE1_MIN<=length && length<=SIZE1_MAX);
        }
    } // switch
    return(true);
}

// .................................................................

static parse_error_t parse_string(uint16_t *str_num, str_list_t **str_list, uint16_t option_number,
                            uint8_t *input, uint16_t length) {

    if (!option_length_ok(option_number, length)) {
        return(PE_OPTION_OUT_OF_BOUNDS);
    }
    
    str_list_t *new_val = malloc(sizeof(str_list_t));
    if (new_val == NULL) {
        return(PE_OUT_OF_MEMORY);
    }
    str_opt_t *str_val = malloc(sizeof(str_opt_t));
    if (str_val == NULL) {
        free((void*)new_val);
        return(PE_OUT_OF_MEMORY);
    }
    if (length==0) {
        str_val->string=NULL;
    } else {
        char *tmp_str = malloc(length);
        if (tmp_str == NULL) {
            free((void*)new_val);
            free((void*)str_val);
            return(PE_OUT_OF_MEMORY);
        }
        (void)memcpy(tmp_str,input,length);
        str_val->string=tmp_str;
    }
    str_val->length = length;
    str_val->number = option_number;
    
    new_val->string = str_val;
    new_val->next   = *str_list;
    *str_list       = new_val;
    ++(*str_num);
    return(PE_OK);
}

// .................................................................

static parse_error_t parse_opaque(uint16_t *opaq_num, opaq_list_t **opaq_list, uint16_t option_number,
                            uint8_t *input, uint16_t length) {

    if (!option_length_ok(option_number, length)) {
        return(PE_OPTION_OUT_OF_BOUNDS);
    }
    
    opaq_list_t *new_val = malloc(sizeof(opaq_list_t));
    if (new_val == NULL) {
        return(PE_OUT_OF_MEMORY);
    }
    opaq_opt_t *opaq_val = malloc(sizeof(opaq_opt_t));
    if (opaq_val == NULL) {
        free((void*)new_val);
        return(PE_OUT_OF_MEMORY);
    }
    if (length==0) {
        opaq_val->opaque=NULL;
    } else {
        uint8_t *tmp_opaq = malloc(length);
        if (tmp_opaq == NULL) {
            free((void*)new_val);
            free((void*)opaq_val);
            return(PE_OUT_OF_MEMORY);
        }
        (void)memcpy(tmp_opaq,input,length);
        opaq_val->opaque=tmp_opaq;
    }
    opaq_val->length = length;
    opaq_val->number = option_number;
    
    new_val->opaque = opaq_val;
    new_val->next   = *opaq_list;
    *opaq_list       = new_val;
    ++(*opaq_num);
    return(PE_OK);
}

// .................................................................

static parse_error_t parse_uint(uint16_t *uint_num, uint_list_t **uint_list, uint16_t option_number,
                            uint8_t *input, uint16_t length) {

    if (length > 8) return(PE_UINT_GT64);
    
    if (!option_length_ok(option_number, length)) {
        return(PE_OPTION_OUT_OF_BOUNDS);
    }
    
    uint_list_t *new_val = malloc(sizeof(uint_list_t));
    if (new_val == NULL) {
        return(PE_OUT_OF_MEMORY);
    }
    uint_opt_t *uint_val = malloc(sizeof(uint_opt_t));
    if (uint_val == NULL) {
        free((void*)new_val);
        return(PE_OUT_OF_MEMORY);
    }
    if (length==0) {
        uint_val->uint_val=0;
    } else {
        swap64_t rotate_bytes;
        rotate_bytes.value = 0;
        unsigned offset = length;
        for (unsigned i=0; i<length; ++i,--offset) {
            rotate_bytes.bytes[offset] = input[i];
        }
        uint_val->uint_val = rotate_bytes.value;
    }
    uint_val->number  = option_number;

    new_val->uint_val = uint_val;
    new_val->next     = *uint_list;
    *uint_list        = new_val;
    ++(*uint_num);
    return(PE_OK);
}

// .................................................................

static parse_error_t parse_empty(uint16_t *empty_num, empty_list_t **empty_list, uint16_t option_number) {
    empty_list_t *new_val = malloc(sizeof(empty_list_t));
    if (new_val == NULL) {
        return(PE_OUT_OF_MEMORY);
    }
    new_val->value = option_number;
    new_val->next  = *empty_list;
    *empty_list    = new_val;
    ++(*empty_num);
    return(PE_OK);
}


// EOF parse_coap
// .................................................................
// .................................................................
