/*
 * Author: Dušan Kolář
 */

#include <stdio.h>
#include <malloc.h>

#include "pcx.h"

void demo_parse_packet(uint8_t *data, long packet_size) {
    puts("Parsing...");
    
    parse_error_t error;
    coap_data_t *coap = parse_coap(data, (uint32_t)packet_size, &error);
    
    puts("Parsing completed...");
    
    if (coap == NULL) {
        puts("Packet parsing failed");
        switch (error) {
            case PE_TOO_SHORT: {
                puts("PE_TOO_SHORT");
                break;
            }
            case PE_OUT_OF_DATA: {
                puts("PE_OUT_OF_DATA");
                break;
            }
            case PE_WRONG_PAYLOAD_MARKER: {
                puts("PE_WRONG_PAYLOAD_MARKER");
                break;
            }
            case PE_EXPECTED_VERSION_01: {
                puts("PE_EXPECTED_VERSION_01");
                break;
            }
            case PE_TOKEN_LENGTH_SIZE: {
                puts("PE_TOKEN_LENGTH_SIZE");
                break;
            }
            case PE_ZERO_LENGTH_PAYLOAD: {
                puts("PE_ZERO_LENGTH_PAYLOAD");
                break;
            }
            case PE_OUT_OF_MEMORY: {
                puts("PE_OUT_OF_MEMORY");
                break;
            }
            case PE_OPTION_OUT_OF_BOUNDS: {
                puts("PE_OPTION_OUT_OF_BOUNDS");
                break;
            }
            case PE_UINT_GT64: {
                puts("PE_UINT_GT64");
                break;
            }
            default: {
                puts("Unknown reason");
                break;
            }
        }
    } else {
        printf("Type: %u\n", coap->type);
        printf("Code: %02x\n", coap->code);
        printf("ID: %04x\n", coap->id);
        if (coap->tkl==0) {
            puts("Token: NONE");
        } else {
            fputs("Token: ",stdout);
            for (unsigned i=0; i<coap->tkl; ++i) {
                printf("%02x",coap->token[i]);
            }
            puts("");
        }
        printf("Options: %u\n", coap->options_num);
        printf("String options: %u\n", coap->str_num);
        if (coap->str_num > 0) {
            for (unsigned i=0; i<coap->str_num; ++i) {
                str_opt_t *val = coap->strings[i];
                printf("  string %3u val: ",val->number);
                for (unsigned j=0; j<val->length; ++j) {
                    putchar(val->string[j]);
                }
                puts("");
            }
        }
        printf("Opaque options: %u\n", coap->opaq_num);
        if (coap->opaq_num > 0) {
            for (unsigned i=0; i<coap->opaq_num; ++i) {
                opaq_opt_t *val = coap->opaques[i];
                printf("  opaque %3u val: ",val->number);
                for (unsigned j=0; j<val->length; ++j) {
                    printf("%02x",val->opaque[j]);
                }
                puts("");
            }
        }
        printf("UInt options: %u\n", coap->uint_num);
        if (coap->uint_num > 0) {
            for (unsigned i=0; i<coap->uint_num; ++i) {
                uint_opt_t *val = coap->uints[i];
                printf("  uint %3u val: %10lu\n",val->number,val->uint_val);
            }
        }
        printf("Empty options: %u\n", coap->empty_num);
        if (coap->empty_num > 0) {
            for (unsigned i=0; i<coap->empty_num; ++i) {
                printf("  empty %3u\n",coap->emptys[i]);
            }
        }
        printf("Payload length: %u\n", coap->payload_length);
        if (coap->payload_length > 0) {
            printf("  data:");
            for(unsigned i=0; i<8 && i<coap->payload_length; ++i) {
                uint8_t val = coap->payload[i];
                if (val>=32 && val<127) {
                    printf(" %02x (%c)",val,val);
                } else {
                    printf(" %02x    ",val);
                }
            }
            puts("");
        }
        free_coap_data(coap);
    }
    
}

int main(int argc, char *argv[]) {
    if (argc<2) {
        puts("Usage: test <coap_packet_file(s)>");
        return 1;
    }
    
    for (int i=1; i<argc;++i) {
        printf("Started %s...\n",argv[i]);
    
        FILE *packet;
        if ((packet = fopen(argv[i],"rb")) == NULL) {
            printf("Cannot open file: %s\n", argv[i]);
            return 2;
        }
        
        if (fseek(packet, 0L, SEEK_END)) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 3;
        }
        
        long packet_size = ftell(packet);
        
        if (packet_size<0) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 4;
        }
        
        if (packet_size > 0x10000L) {
            printf("Packet size too large: %ld\n",packet_size);
            fclose(packet);
            return 5;
        }
        
        if (fseek(packet, 0L, SEEK_SET)) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 6;
        }
        
        uint8_t *data = malloc((size_t)packet_size);
        
        if (fread((void*)data,1,(size_t)packet_size,packet) != (size_t)packet_size) {
            printf("Cannot read file: %s\n",argv[i]);
            fclose(packet);
            return 7;
        }
        
        fclose(packet);
        
        demo_parse_packet(data, packet_size);
        
        free((void*)data);
        
        printf("Done %s !\n",argv[i]);
        
    } // end for
    
    return 0;
} // main

// EOF
